/*
** Module   : CompoundQuery.java
** Abstract : Implementation of an iterative, multi-table query
**
** Copyright (c) 2004-2010, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20051112   @23394 Created initial version. Implementation of an
**                           iterative, multi-table query. Component
**                           queries are added and the join of the results
**                           is performed at the client (i.e., P2J server,
**                           but database client).
** 002 ECF 20051216   @23746 Added scrolling support. Now extends the new
**                           DynamicQuery base class, which provides
**                           cursoring support services. Modified
**                           retrieval logic to allow for repositioning.
** 003 ECF 20060104   @23826 Changes to support new hierarchy. Added
**                           unsupported variants of unique() method
**                           required by P2JQuery interface. Both throw
**                           UnsupportedOperationException.
** 004 ECF 20060215   @24715 Added support for client-side where clause
**                           processing. Also refitted to implement more
**                           correct handling of substitution parameter
**                           resolution for iterative looping queries.
** 005 ECF 20060531   @26849 Replaced EndConditionException with
**                           QueryOffEndException. The latter is a more
**                           specific version of the former, used to
**                           indicate a query has run off the end of its
**                           result set.
** 006 ECF 20060705   @27881 Fixed preselect mode of retrieve method.
**                           Invocations of this method after the first
**                           would always take the dynamic route rather
**                           than loading records by IDs.
** 007 ECF 20060801   @28349 Added getBreakValue() method to meet
**                           superclass requirement. Currently throws
**                           UnsupportedOperationException.
** 008 ECF 20060925   @29962 Check for a pending error before retrieving a
**                           record. Retrieval is not attempted if an
**                           error is pending in the ErrorManager.
** 009 ECF 20061003   @30152 Backed out #009 (@29962). This logic is only
**                           necessary for FindQuery and AdaptiveFind.
** 010 ECF 20061006   @30196 Implemented scrolling mode. When set to
**                           scrolling mode, non-linear scrolling is
**                           possible in preselect mode. Previously, only
**                           linear scrolling from first to last was
**                           possible. This is necessitated by the fact
**                           that CompoundQuery may back a browse widget.
**                           Scrolling support is actually implemented in
**                           component queries. At this level, a flag is
**                           simply set to enable the deferred setting of
**                           scrolling mode on component queries.
** 011 ECF 20061019   @30525 Fixes to better support scrolling mode.
**                           Modified the behavior of some state variables
**                           to fix the use of AdaptiveQuery components in
**                           scrolling mode.
** 012 ECF 20061204   @31574 Implemented is/setBrowsed() methods. Browse
**                           association status is propagated to component
**                           queries.
** 013 ECF 20070323   @32548 Fixed ClassCastException in setScrolling().
** 014 ECF 20070504   @33407 Fixed loadByIDs(). If a component query can
**                           not load the given ID set, return false.
** 015 ECF 20070509   @33453 Fixed problem of resetting component queries
**                           too early. Added notifyRepositionListeners()
**                           implementation to avoid sending notifications
**                           before query is fully initialized.
** 016 ECF 20070511   @33471 Fixed loadByIDs(). Reset component queries
**                           before loading records.
** 017 ECF 20070518   @33678 Implement recordBuffers(). Required by the
**                           RecordChangeListener interface, which is now
**                           implemented by the superclass.
** 018 ECF 20070716   @34507 Add a cleanup() method implementation. All
**                           component queries are marked as not being
**                           standalone, so that they do not clean up
**                           their resources too aggressively, but rather
**                           defer to this object's cleanup mechanism.
** 019 ECF 20070726   @34720 Fixed problem of outer buffers being cleared
**                           prematurely on loop iteration. When we begin
**                           a new iteration series, we temporarily
**                           disable RecordBuffer's default release on
**                           iterate behavior, but only for the current
**                           block scope.
** 020 ECF 20070821   @34906 Added support for temporary silent error
**                           mode. This allows query substitution parms to
**                           be processed safely during initialization of
**                           a query.
** 021 ECF 20071011   @35443 Optimization to loadByIDs(). The outer
**                           table's query does not need to be reset when
**                           loading buffers by ID.
** 022 ECF 20071203   @36119 Integrated generics for RecordChangeListener
**                           implementation.
** 023 ECF 20080112   @36997 Removed restriction of dealing only with IDs.
**                           This query can now handle both Persistables
**                           and primary key IDs when retrieving data.
**                           This is necessary when the full record is
**                           needed immediately, not just an ID, such as
**                           in presort cases. Removed final modifier from
**                           class.
** 024 ECF 20080509   @38241 Added open() method. Necessary for converted
**                           OPEN QUERY statements.
** 025 ECF 20080509   @38251 Rolled back #024 (@38241). Caused regression.
** 026 ECF 20080510   @38479 Changed setScrolling(). Do not process method
**                           body more than once.
** 027 SIY 20080625   @38922 Removed duplicate browsed flag which under
**                           some circumstances may be set in opposite
**                           state to identical flag in AbstractQuery
**                           and disable correct browse notifications.
** 028 ECF 20081031   @40306 Minor memory improvement.
** 029 ECF 20081105   @40216 Overrode parent's open() method. Registers
**                           all query components with the ChangeBroker at
**                           the appropriate scope.
** 030 ECF 20081106   @40320 Fixed open(). In case of a scrolling query,
**                           the CompoundQuery itself must be registered
**                           with the ChangeBroker at the proper scope.
** 031 CA  20081210   @40837 Implemented CLOSE QUERY statement.
** 032 ECF 20090223   @41351 Fixed retrieveImpl(). Reset activeIndex on
**                           FIRST/LAST navigation types to ensure we are
**                           fetching all components' results.
** 033 ECF 20090312   @42980 Improved reposition support. Track number of
**                           result rows visited, and which direction we go
**                           off end. Implement loadRowAtCursor() for
**                           scrolling queries.
** 034 ECF 20090703   @43056 Fixed loadByValues(). If an expected record can
**                           not be loaded, MissingRecordException is thrown
**                           by Joinable.load() and the backing buffer is
**                           assumed by contract to be in unknown (safe) mode,
**                           such that a request for the buffer record's
**                           property values will return unknown value.
** 035 ECF 20090716   @43215 Ensure backing record buffers are initialized
**                           before use.
** 036 CA  20090727   @43393 When the query goes off-end, all buffers for this
**                           query are registered for off-end notifications, 
**                           in case of rollback (when current block is FOR
**                           or FOR loop).
** 037 CA  20090731   @43465 Register the query for off-end listener 
**                           registration after block initialization (done
**                           only for block types FOR and FOR EACH). Added
**                           getOffEndListeners, which returns a list of all
**                           off-end listeners for all used buffers.
** 038 SVL 20091030   @44288 getOffEnd() function was made public in order to
**                           support API change into P2JQuery. 
** 039 SIY 20100303   @44702 Allow enablind/disabling of accumulator 
**                           processing.
** 040 VMN 20130522          ErrorIfNull is used instead of boolean and method unique() returns
**                           boolean.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import org.apache.commons.logging.*;
import com.goldencode.p2j.util.*;

/**
 * A client-side (relative to the database server) implementation of a
 * multi-table, join query.  Queries of individual tables are added via one
 * of the <code>addComponent</code> method variants.  This query type can
 * combine both {@link PreselectQuery}s and {@link RandomAccessQuery}s as its
 * components.  However, preselected tables must always be to the left of
 * (i.e., outward from) non-preselected tables in the query.  That is, a
 * <code>PreselectQuery</code> object can never be added, once a
 * <code>RandomAccessQuery</code> object has been added to a compound query.
 * <p>
 * Inner and left outer joins are supported.  However, inner joins must
 * always occur to the left (i.e., outward from) outer joins.  If a component
 * query specified as an inner join is added <i>after</i> any previously
 * added component query specified as an outer join will automatically be
 * converted to an inner join.
 * <p>
 * This query retrieves result records in two modes:  iterating and random
 * access.  Iterating mode is active when using the {@link #iterate} method
 * to advance the query to the next, composite row of records in the result
 * set.  Random access mode is active when using any of the other retrieval
 * methods (<code>first</code>, <code>last</code>, <code>next</code>,
 * <code>previous</code>, <code>current</code>).  The difference between the
 * modes is that iterating mode always terminates with a
 * <code>QueryOffEndException</code> when no further record retrieval is
 * possible;  random access mode does not.  Random access mode thus allows
 * client code to advance off the end of the result set and continue
 * processing.
 * <p>
 * <code>CompoundQuery</code> abstracts the use of multiple query
 * implementations to achieve a dynamic, multi-table query.  It delegates the
 * record retrieval work to the underlying queries, but manages the overall
 * navigation of the results.  Delegate queries must implement the {@link
 * Joinable} interface.  Each "row" in the "result set" is a logical
 * construct, the virtual composite of multiple record buffers.  Advancing to
 * a new row (be it forwards, backwards, first, last) means advancing the
 * most specific query possible to the next matching record in its
 * predetermined progression.  Once the most specific query can no longer
 * advance, the next most specific query is advanced, the most specific query
 * is reset to its starting state, and we try to find a joining match for the
 * most specific query again, given that its outer query has a new value with
 * which to join.  This continues until the most specific possible query has
 * advanced, or we have moved to our least specific query component, and no
 * further match is possible.  At this point, the query is "off-end".  For
 * iterating mode, this is the end of query execution and an end condition is
 * raised.  In random access mode, further navigation is possible.
 * <p>
 * <b>Substitution Parameters</b><br>
 * Query substitution parameters require different modes of resolution,
 * depending upon the type of parameter and the nature of the query.  A query
 * may be set at construction to resolve all of its parameters once, up front
 * or iteratively, each time an iterative loop construct is entered.  The
 * former corresponds with Progress' OPEN QUERY style queries;  the latter
 * corresponds with all other iterative record retrieval constructs in
 * Progress (e.g., FOR EACH, DO PRESELECT, REPEAT PRESELECT, REPEAT FOR).
 * <p>
 * The number of times a parameter is resolved and the timing of such
 * resolution becomes important if an iterative record retrieval construct
 * modifies the values of such parameters during record retrieval.  This may
 * occur, for example, if the body of an outer FOR loop modifies the value of
 * a variable used in the where clause of an inner FOR loop.  While less of
 * an issue for preselect type queries, these can still be affected if the
 * where clause of a nested query component invokes a client-side function,
 * which may in turn modify state on which the query or related logic depends.
 * <p>
 * The following conversion rules apply:
 * <pre>
 * -----------------------------------------------------------------
 * Progress Construct                     CompoundQuery Modes
 * -----------------------------------------------------------------
 * OPEN QUERY FOR                         resolveOnce = true
 *                                        preselect   = false
 * OPEN QUERY PRESELECT                   resolveOnce = true
 *                                        preselect   = true
 * FOR {EACH | FIRST | LAST}              resolveOnce = false
 *                                        preselect   = false
 * DO PRESELECT                           resolveOnce = false
 *                                        preselect   = true
 * REPEAT PRESELECT                       resolveOnce = false
 *                                        preselect   = true
 * -----------------------------------------------------------------
 * </pre>
 */
public class CompoundQuery
extends DynamicQuery
implements QueryConstants
{
   /** Logger */
   private static final Log LOG = Utils.getLog(CompoundQuery.class);
   
   /** List of components */
   private final List<Component> components = new ArrayList<Component>(2);
   
   /** Resolve arguments once, up front, or on every loop cycle? */
   private final boolean resolveOnce;
   
   /** Force preselection of all results? */
   private final boolean preselect;
   
   /** Flag indicating whether component queries should be scrollable */
   private boolean scrolling = false;
   
   /** Have results been preselected? */
   private boolean preselected = false;
   
   /** Current index position in each component query loop */
   private int[] counters = null;
   
   /** Index of component query currently being iterated */
   private int activeIndex = -1;
   
   /** Total number of tables involved in this compound query */
   private int tableCount = 0;
   
   /** Postpone accumulation on retrieve. */
   private boolean postponeAccumulation = false;
   
   /**
    * Constructor.
    *
    * @param   resolveOnce
    *          <code>true</code> if query substitution arguments are to be
    *          resolved once, up front;  <code>false</code> if arguments must
    *          be resolved for each nested loop cycle.
    * @param   preselect
    *          <code>true</code> to force preselection of all results;
    *          <code>false</code> otherwise.
    */
   public CompoundQuery(boolean resolveOnce, boolean preselect)
   {
      this.resolveOnce = resolveOnce;
      this.preselect = preselect;

      TransactionManager.registerOffEndQuery(this);
   }
   
   /**
    * Create a cursor for this query and ensure that any component queries
    * are scrollable, whether they already have been added, or are added in
    * the future.
    */
   public void setScrolling()
   {
      if (!scrolling)
      {
         scrolling = true;
         super.setScrolling();
         
         Iterator<Component> iter = components.iterator();
         while (iter.hasNext())
         {
            iter.next().getQuery().setScrolling();
         }
      }
   }
   
   /**
    * Get the number of tables joined by this query.
    * 
    * @return  Number of tables joined by this query.
    */
   public int getTableCount()
   {
      return tableCount;
   }

   /**
    * Get all the off-end listeners associated with this query.
    * 
    * @return  the off-end listeners associated with this query.
    */
   public List<QueryOffEndListener> getOffEndListeners()
   {
      List<QueryOffEndListener> listeners = 
         new ArrayList<QueryOffEndListener>(tableCount);
      
      for (int i = 0; i < components.size(); i++)
      {
         listeners.addAll(components.get(i).getQuery().getOffEndListeners()); 
      }
      
      return listeners;
   }
   
   /**
    * Add a joinable query to the compound query.  If this is not the first
    * component to be added to the query, its results will be joined to the
    * immediately previous component by an inner join.  This query will be
    * iterated using the <code>NEXT</code> iteration type.
    *
    * @param   query
    *          Query which will perform the fundamental work for this
    *          component when the compound query is iterated.
    *
    * @throws  IllegalStateException
    *          if this method is invoked during an iteration cycle.
    *
    * @see     #iterate
    */
   public void addComponent(Joinable query)
   {
      addComponent(query, NEXT, false);
   }
   
   /**
    * Add a joinable query to the compound query, and specify the nature of
    * the iteration which will be performed (first, last, next, etc.) on that
    * component, when the compound query is iterated.  If this is not the
    * first component to be added to the query, its results will be joined to
    * the immediately previous component by an inner join.
    *
    * @param   query
    *          Query which will perform the fundamental work for this
    *          component when the compound query is iterated.
    * @param   iteration
    *          A {@link QueryConstants constant} indicating the type of work
    *          to be performed upon an iteration of the compound query.  This
    *          parameter determines whether the underlying query will be asked
    *          to perform a first, last, next, previous, or unique record
    *          retrieval.
    *
    * @throws  IllegalStateException
    *          if this method is invoked during an iteration cycle.
    *
    * @see     #iterate
    */
   public void addComponent(Joinable query, int iteration)
   {
      addComponent(query, iteration, false);
   }
   
   /**
    * Add a joinable query to the compound query, and specify the nature of
    * the iteration which will be performed (first, last, next, etc.) on that
    * component, when the compound query is iterated.  If this is not the
    * first component to be added to the query, the type of join performed
    * with the immediately previous component is determined by the value of
    * <code>outer</code>.
    *
    * @param   query
    *          Query which will perform the fundamental work for this
    *          component when the compound query is iterated.
    * @param   iteration
    *          A {@link QueryConstants constant} indicating the type of work
    *          to be performed upon an iteration of the compound query.  This
    *          parameter determines whether the underlying query will be asked
    *          to perform a first, last, next, previous, or unique record
    *          retrieval.
    * @param   outer
    *          <code>true</code> if this component should be joined to the
    *          previously added component via a left outer join;
    *          <code>false</code> if an inner join is required.  Note that
    *          adding an inner join component will cause any previously added
    *          components added as outer joins to be reset to inner joins.
    *
    * @throws  IllegalStateException
    *          if this method is invoked during an iteration cycle.
    *
    * @see     #iterate
    */
   public void addComponent(Joinable query, int iteration, boolean outer)
   {
      if (activeIndex >= 0)
      {
         throw new IllegalStateException(
            "Cannot add query component while iterating");
      }
      
      // Clear temporary silent error mode.
      init();
      
      // Mark component query as not standalone.
      query.setStandalone(false);
      
      // In a compound query, joined first and last record retrievals must
      // not throw an ErrorConditionException if no record is found.
      query.setErrorIfNull(ErrorIfNull.NO_ERROR);
      
      // An inner join effectively makes all previous outer joins into inner
      // joins, so make this switch implicit.
      // TODO:  log warning to this effect.
      if (!outer)
      {
         ListIterator<Component> iter =
            components.listIterator(components.size());
         while (iter.hasPrevious())
         {
            Component comp = iter.previous();
            comp.setInner();
         }
      }
      
      // Does the component query need to scroll?
      if (scrolling)
      {
         query.setScrolling();
      }
      
      // If necessary, indicate to the component query that it should behave
      // as if it were associated with a browse.
      if (isBrowsed())
      {
         query.setBrowsed(isBrowsed());
      }
      
      components.add(new Component(query, iteration, outer));
      
      // Update total table count.
      tableCount += query.getTableCount();
   }
   
   /**
    * Indicate to this query that whether it is associated with a browse
    * widget or not.
    *
    * @param   browsed
    *          <code>true</code> to indicate the query has an associated
    *          browse widget, else <code>false</code>.
    */
   public void setBrowsed(boolean browsed)
   {
      super.setBrowsed(browsed);
      
      Iterator<Component> iter = components.iterator();
      while (iter.hasNext())
      {
         Component next = iter.next();
         next.getQuery().setBrowsed(browsed);
      }
   }
   
   /**
    * Open the query once all query components have been added.
    */
   public void open()
   {
      // Find for each compound query component the most outer scope among
      // nearest external scope and the outermost scopes of buffers used by
      // this component and register the component into this scope.
      int externalScope = TransactionManager.findNearestExternal();
      int lowestScope = externalScope;
      
      for (Component comp : components)
      {
         Joinable query = comp.getQuery();
         int scope = externalScope;
         RecordBuffer[] buffers = query.getRecordBuffers();

         for (RecordBuffer buffer : buffers)
         {
            scope = ((scope == -1)
                     ? buffer.getScopeOpenDepth()
                     : Math.min(scope, buffer.getScopeOpenDepth()));
            lowestScope = ((lowestScope == -1)
                           ? scope
                           : Math.min(lowestScope, scope));
         }

         if (scope != -1)
         {
            query.registerRecordChangeListeners(scope);
         }
      }
      
      super.open();
   }
   
   /**
    * Advance the compound query, triggering one or more of the underlying
    * queries to retrieve a new record.  This produces a virtual, composite
    * row of results which is the combination of the records retrieved by
    * each component query.
    *
    * @throws  QueryOffEndException
    *          if no more records can be retrieved by the underlying queries.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void iterate()
   {
      retrieve(NEXT, null, true);
   }
   
   /**
    * Retrieve the first virtual, composite row of results for the compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void first()
   {
      first(false, null);
   }
   
   /**
    * Retrieve the first virtual, composite row of results for the compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void first(boolean iterating)
   {
      first(iterating, null);
   }
   
   /**
    * Retrieve the first virtual, composite row of results for the compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void first(LockType lockType)
   {
      first(false, lockType);
   }
   
   /**
    * Retrieve the first virtual, composite row of results for the compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void first(boolean iterating, LockType lockType)
   {
      boolean scrolling = (cursor != null);
      boolean done = false;
      
      if (scrolling)
      {
         done = loadByValues(cursor.getFirst(), lockType, iterating);
      }
      
      if (!done)
      {
         Object[] ids = retrieve(FIRST, lockType, iterating);
         if (scrolling)
         {
            cursor.addResultFirst(ids);
         }
      }
   }
   
   /**
    * Retrieve the last virtual, composite row of results for the compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void last()
   {
      last(false, null);
   }
   
   /**
    * Retrieve the last virtual, composite row of results for the compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void last(boolean iterating)
   {
      last(iterating, null);
   }
   
   /**
    * Retrieve the last virtual, composite row of results for the compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void last(LockType lockType)
   {
      last(false, lockType);
   }
   
   /**
    * Retrieve the last virtual, composite row of results for the compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void last(boolean iterating, LockType lockType)
   {
      boolean scrolling = (cursor != null);
      boolean done = false;
      
      if (scrolling)
      {
         done = loadByValues(cursor.getLast(), lockType, iterating);
      }
      
      if (!done)
      {
         Object[] ids = retrieve(LAST, lockType, iterating);
         if (scrolling)
         {
            cursor.addResultLast(ids);
         }
      }
   }
   
   /**
    * Retrieve the next virtual, composite row of results for the compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void next()
   {
      next(false, null);
   }
   
   /**
    * Retrieve the next virtual, composite row of results for the compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void next(boolean iterating)
   {
      next(iterating, null);
   }
   
   /**
    * Retrieve the next virtual, composite row of results for the compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void next(LockType lockType)
   {
      next(false, lockType);
   }
   
   /**
    * Retrieve the next virtual, composite row of results for the compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void next(boolean iterating, LockType lockType)
   {
      boolean scrolling = (cursor != null);
      boolean done = false;
      
      if (scrolling)
      {
         Object[] data;
         while (!done && (data = cursor.getNext()) != null)
         {
            done = loadByValues(data, lockType, iterating);
         }
      }
      
      if (!done)
      {
         Object[] ids = retrieve(NEXT, lockType, iterating);
         if (scrolling)
         {
            cursor.addResultNext(ids);
         }
      }
   }
   
   /**
    * Retrieve the previous virtual, composite row of results for the compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void previous()
   {
      previous(false, null);
   }
   
   /**
    * Retrieve the previous virtual, composite row of results for the compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void previous(boolean iterating)
   {
      previous(iterating, null);
   }
   
   /**
    * Retrieve the previous virtual, composite row of results for the compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void previous(LockType lockType)
   {
      previous(false, lockType);
   }
   
   /**
    * Retrieve the previous virtual, composite row of results for the compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void previous(boolean iterating, LockType lockType)
   {
      boolean scrolling = (cursor != null);
      boolean done = false;
      
      if (scrolling)
      {
         Object[] data;
         while (!done && (data = cursor.getPrevious()) != null)
         {
            done = loadByValues(data, lockType, iterating);
         }
      }
      
      if (!done)
      {
         Object[] ids = retrieve(PREVIOUS, lockType, iterating);
         if (scrolling)
         {
            cursor.addResultPrevious(ids);
         }
      }
   }
   
   /**
    * Retrieve the current virtual, composite row of results for the compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void current()
   {
      current(null);
   }
   
   /**
    * Retrieve the current virtual, composite row of results for the compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    */
   public void current(LockType lockType)
   {
      retrieve(CURRENT, lockType, false);
   }
   
   /**
    * Retrieve a unique record (only valid for a single-table query).
    * <p>
    * This method is required by the <code>P2JQuery</code> interface, but it
    * is invalid for a compound query.
    *
    * @return  <code>true</code> if no errors occurred during record retrieval and exactly one
    *          record is found which match the criteria else <code>false</code>.
    *
    * @throws  UnsupportedOperationException
    *          always.
    */
   public boolean unique()
   {
      return unique(null);
   }
   
   /**
    * Retrieve a unique record (only valid for a single-table query),
    * overriding the default lock type of the underlying query component.
    * <p>
    * This method is required by the <code>P2JQuery</code> interface, but it
    * is invalid for a compound query.
    *
    * @return  <code>true</code> if no errors occurred during record retrieval and exactly one
    *          record is found which match the criteria else <code>false</code>.
    *
    * @param   lockType
    *          Lock type which should override that of the query component.
    *
    * @throws  UnsupportedOperationException
    *          always.
    */
   public boolean unique(LockType lockType)
   {
      throw new UnsupportedOperationException();
   }
   
   /**
    * Fetch the array of primary key IDs associated with the first result in
    * this query's result list.
    * <p>
    * This implementation actually updates the underlying record buffer(s).
    *
    * @return  Array of record IDs reflecting the first query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekFirst()
   {
      return retrieve(FIRST, LockType.NONE, false);
   }
   
   /**
    * Fetch the array of primary key IDs associated with the last result in
    * this query's result list.
    * <p>
    * This implementation actually updates the underlying record buffer(s).
    *
    * @return  Array of record IDs reflecting the last query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekLast()
   {
      return retrieve(LAST, LockType.NONE, false);
   }
   
   /**
    * Fetch the array of primary key IDs associated with the next result in
    * this query's result list.
    * <p>
    * This implementation actually updates the underlying record buffer(s).
    *
    * @return  Array of record IDs reflecting the next query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekNext()
   {
      return retrieve(NEXT, LockType.NONE, false);
   }
   
   /**
    * Fetch the array of primary key IDs associated with the previous result
    * in this query's result list.
    * <p>
    * This implementation actually updates the underlying record buffer(s).
    *
    * @return  Array of record IDs reflecting the previous query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekPrevious()
   {
      return retrieve(PREVIOUS, LockType.NONE, false);
   }
   
   /**
    * Notify all registered listeners that query is repositioned or closed.  
    * If this is invoked before any query components have been added to the 
    * compound query, it is ignored.
    * 
    * @param   closed
    *          A <code>true</code> value will inform all listeners that the
    *          query was explicitly closed.
    */
   public void notifyRepositionListeners(boolean closed)
   {
      if (!components.isEmpty())
      {
         super.notifyRepositionListeners(closed);
      }
   }
   
   /**
    * Report the record buffers for whose changes this object is interested
    * in listening.  This listener will be registered with the {@link
    * ChangeBroker} to receive notifications of any change to DMOs whose
    * interface types match those returned by the {@link
    * RecordBuffer#getDMOInterface} methods of the returned buffers.
    *
    * @return  An iterator on the record buffers stored in each of this
    *          query's components.
    */
   public Iterator<RecordBuffer> recordBuffers()
   {
      Iterator<RecordBuffer> iter = new Iterator<RecordBuffer>()
      {
         private Iterator<Component> compIter = components.iterator();
         
         private RecordBuffer nextBuffer = null;
         
         private RecordBuffer[] bufs = null;
         
         private int index = -1;
         
         public boolean hasNext()
         {
            if (nextBuffer == null)
            {
               advance();
            }
            
            return (nextBuffer != null);
         }
         
         public RecordBuffer next()
         {
            if (!hasNext())
            {
               throw new NoSuchElementException();
            }
            
            try
            {
               return nextBuffer;
            }
            finally
            {
               nextBuffer = null;
            }
         }
         
         public void remove()
         {
            throw new UnsupportedOperationException();
         }
         
         private void advance()
         {
            if (bufs != null)
            {
               if (++index < bufs.length)
               {
                  nextBuffer = bufs[index];
                  
                  return;
               }
               
               bufs = null;
            }
            
            if (bufs == null)
            {
               if (compIter.hasNext())
               {
                  Component nextComp = compIter.next();
                  Joinable query = nextComp.getQuery();
                  bufs = query.getRecordBuffers();
                  index = -1;
               }
               else
               {
                  nextBuffer = null;
                  
                  return;
               }
            }
            
            // Assumes that array will not be empty or null.
            nextBuffer = bufs[++index];
         }
      };
      
      return iter;
   }

   /**
    * Clean up resources associated with this query when its scope ends.
    */
   public void cleanup()
   {
      if (closed)
      {
         return;
      }

      Iterator<Component> iter = components.iterator();
      while (iter.hasNext())
      {
         Component next = iter.next();
         Joinable query = next.getQuery();
         query.cleanup();
      }
   }
   
   /**
    * Report the off-end status of this query, which is essentially the
    * off-end status of its outermost component query.  The query is not
    * considered off-end in its starting state;  that is, if a result has not
    * yet been retrieved.
    * 
    * @return  One of the {@link OffEnd} constants <code>NONE</code>,
    *          <code>FRONT</code>, or <code>BACK</code>.
    */
   public OffEnd getOffEnd()
   {
      AbstractQuery query = (AbstractQuery) components.get(0).getQuery();
      
      return query.getOffEnd();
   }
   
   /**
    * Return an object which represents a property value which defines the
    * end of a <i>sort band</i>.  A <i>sort band</i> is a series of one or
    * more records which share the same value for a particular property,
    * where that property is used as the coarsest sorting criterion defined
    * by this query's <code>order by</code> clause.
    * <p>
    * If the break value for this query currently is non-<code>null</code>,
    * this indicates that the most recently retrieved record has crossed a
    * <i>sort band</i> boundary, and that a new band has begun.  This value
    * will only be non-<code>null</code> at the point at which such a record
    * is found, and will be <code>null</code> for all other records.
    * <p>
    * Note:  break values are only tracked for invocations of
    * <code>next</code> and <code>previous</code>.  Other retrievals will
    * result in break value being <code>null</code>.
    *
    * @return  No value is returned;  this operation is not supported for
    *          this query type.
    *
    * @throws  UnsupportedOperationException
    *          always.
    */
   protected Object getBreakValue()
   {
      // TODO:  implement this method to return a meaningful value.
      throw new UnsupportedOperationException();
   }
   
   /**
    * Retrieve a virtual, composite row of results for the compound query.
    * This method implements a client-side join of multiple, single-row
    * record retrievals performed by the underlying, component queries.
    * It works first from outermost component inward, retrieving a record at
    * each level.  When each component's backing buffer contains a record,
    * the composite retrieval is complete and the method returns.  On
    * subsequent iterations, the innermost component is advanced first.  If
    * no more matches are possible at this level, the algorithm moves outward
    * one level and attempts to advance that level.  On success, it moves
    * inward again, on failure, it moves outward again.  This continues until
    * either a new composite row of results is assembled, or no further
    * records can be found.  This progression ensures that each inner query
    * cycles through all of its matches before the next outer query advances.
    * <p>
    * <b>Preselect Mode</b><br>
    * <code>CompoundQuery</code> generally operates as a dynamic query,
    * though it can be used to generate a fixed list of preselected results.
    * When in this mode, the first time this method is executed, a results
    * list is generated.  This is done by retrieving each virtual, composite
    * row which meets the overall query's criteria, and caching the primary
    * keys of these records in the underlying cursor.  Each underlying query
    * is assigned the correct subset of these IDs, such that subsequent
    * retrieval requests can safely use the cursor to retrieve the correct
    * records.  As a result, in preselect mode, this method is only ever
    * invoked once per query. 
    *
    * @param   navigation
    *          Type of advancement applied to the compound query;  i.e., how
    *          the virtual composite result set is navigated as a whole:
    *          forward (next), backward (previous), jump to first or last, or
    *          reload current records in all buffers.
    * @param   lockType
    *          Overriding lock type to apply to each underlying buffer.
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @return  An array of the primary key IDs for every table involved in
    *          the query, (which together identify a unique, virtual record),
    *          if a result row was successfully retrieved;
    *          <code>null</code> if no more results were available (and we
    *          were not iterating).
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an error occurred retrieving the record for an underlying
    *          query.
    */
   @SuppressWarnings("unchecked")
   protected Object[] retrieve(int navigation,
                               LockType lockType,
                               boolean iterating)
   {
      Object[] rowData = null;
      
      if (preselect)
      {
         if (!preselected)
         {
            // Create a cursor.
            setScrolling();
            
            int size = components.size();
            
            // Logging setup.
            boolean trace = LOG.isTraceEnabled();
            StringBuffer buf = null;
            String nl = null;
            if (trace)
            {
               nl = System.getProperty("line.separator");
               buf = new StringBuffer("Preselected IDs:");
            }
            
            // Prepare per-table, preliminary result sets.
            Set[] idSets = new Set[size];
            for (int i = 0; i < size; i++)
            {
               idSets[i] = new LinkedHashSet();
            }
            
            int[] widths = new int[size];
            Iterator<Component> iter = components.iterator();
            for (int i = 0; iter.hasNext(); i++)
            {
               widths[i] = iter.next().getQuery().getTableCount();
            }
            
            // Preselect all records which match the query's criteria.
            do
            {
               rowData = retrieveImpl(NEXT, lockType, false);
               cursor.addResultNext(rowData);
               
               if (rowData != null)
               {
                  int index = 0;
                  for (int i = 0; i < size; i++)
                  {
                     // How many tables does this joinable represent?
                     int w = widths[i];
                     
                     // Initialize the sub-array for this joinable component.
                     // We use an ArrayList instead of an array directly so we
                     // can add this to a Set.
                     List<Object> subIDs = new ArrayList<Object>();
                     for (int j = 0;
                          j < w && index < tableCount;
                          j++, index++)
                     {
                        subIDs.add(rowData[index]);
                     }
                     idSets[i].add(subIDs);
                  }
                  
                  if (trace)
                  {
                     buf.append(nl);
                     buf.append("  ");
                     for (int i = 0; i < size; i++)
                     {
                        buf.append(" ");
                        buf.append(rowData[i]);
                     }
                  }
               }
            }
            while (rowData != null);
            
            // Set up and install final result sets into underlying queries.
            // This is necessary because inner tables in a multi-table query
            // will have an incorrect result set left over from the first pass
            // scan for the last iteration of their enclosing loop.  This
            // would wreak havoc when trying to load from IDs later.
            for (int i = 0; i < size; i++)
            {
               List<Object[]> rows = new ArrayList<Object[]>();
               Iterator<List<Object>> iter2 = idSets[i].iterator();
               while (iter2.hasNext())
               {
                  List<Object> subIDs = iter2.next();
                  rows.add(subIDs.toArray());
               }
               
               SimpleResults results = new SimpleResults(rows);
               Joinable query = components.get(i).getQuery();
               query.setResults(results);
            }
            
            // Reset state and cursor.
            activeIndex = -1;
            preselected = true;
            cursor.beforeFirst();
            
            if (trace)
            {
               LOG.trace(buf.toString());
            }
         }
         
         // Fulfill the retrieval request using the cached results.
         switch (navigation)
         {
            case NEXT:
               rowData = cursor.getNext();
               break;
            case FIRST:
               rowData = cursor.getFirst();
               break;
            case PREVIOUS:
               rowData = cursor.getPrevious();
               break;
            case LAST:
               rowData = cursor.getLast();
               break;
            default:
               rowData = null;
         }
         
         loadByValues(rowData, lockType, iterating);
      }
      else
      {
         // Non-preselect mode;  retrieve composite record dynamically.
         rowData = retrieveImpl(navigation, lockType, iterating);
      }
      
      return rowData;
   }
   
   /**
    * Synchronize the query with the cursor by causing the query to load the
    * current row's results (if any) into its backing buffer(s).
    */
   protected void loadRowAtCursor()
   {
      Object[] row = cursor.getRow();
      if (row != null)
      {
         loadByValues(row, null, false);
      }
   }
   
   /**
    * Load a virtual, composite record by loading each record buffer managed
    * by this query with a DMO, given its primary key ID in the specified
    * <code>ids</code> array.  The IDs are arranged to coincide with the
    * tables underlying the query, from left-most to right-most (in the sense
    * of how they are joined by the query).
    *
    * @param   data
    *          Array of primary key IDs or DMOs, one per table involved in the
    *          query
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *          Set to <code>null</code> to allow underlying query to use its
    *          existing lock type.
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @return  <code>true</code> if <code>ids</code> was not
    *          <code>null</code>;  else <code>false</code>.
    *
    * @see     #collectRowData
    */
   protected boolean loadByValues(Object[] data,
                                  LockType lockType,
                                  boolean iterating)
   {
      if (data == null)
      {
         if (iterating)
         {            
            throw new QueryOffEndException("", 
                  getOffEndListeners().toArray(new QueryOffEndListener[0]));
         }
         
         return false;
      }
      
      // Reset active index and counters.
      activeIndex = 0;
      counters = new int[components.size()];
      
      // Iterate over the query components, updating the buffers of each
      // with the appropriate record.
      Iterator<Component> iter = components.iterator();
      int index = 0;
      for (int i = 0; iter.hasNext(); i++)
      {
         // Get the next query component.
         Component comp = iter.next();
         
         // Get the underlying query.
         Joinable query = comp.getQuery();
         
         // How many tables does this joinable query represent?
         int width = query.getTableCount();
         
         // Initialize the sub-array for this joinable component.
         Object[] subData = new Object[width];
         for (int j = 0; j < width && index < tableCount; j++, index++)
         {
            subData[j] = data[index];
         }
         
         // Update the component query with the proper record(s).
         if (i > 0)
         {
            query.reset(!resolveOnce);
         }
         
         try
         {
            query.load(subData, lockType);
         }
         catch (MissingRecordException exc)
         {
            // The expected record could not be loaded, so the backing buffer
            // is now empty (and should be in safe, unknown mode).
            continue;
         }
         catch (PersistenceException exc)
         {
            if (LOG.isWarnEnabled())
            {
               LOG.warn("Error loading data", exc);
            }
            
            return false;
         }
         catch (QueryOffEndException exc)
         {
            return false;
         }
         catch (ConditionException exc)
         {
            activeIndex = -1;
            
            throw exc;
         }
         
         // Reflect the fact that this component has been updated.
         counters[i]++;
         activeIndex++;
      }
      
      activeIndex--;
      
      return true;
   }
   
   /**
    * Assemble an array of primary key IDs or DMOs for the current records in
    * the buffers underlying this query, from left-most to right-most (in the
    * sense of how the query joins the associated tables).  This method
    * essentially takes a lightweight snapshot of the currently loaded
    * records, so that this may be restored later.
    * <p>
    * This snapshot is only generated for scrolling queries.
    *
    * @param   navigation
    *          Type of advancement applied to the compound query;  i.e., how
    *          the virtual composite result set is navigated as a whole:
    *          forward (next), backward (previous), jump to first or last, or
    *          reload current records in all buffers.
    * 
    * @return  Array of primary key IDs or DMOs, one per table involved in the
    *          query.
    *
    * @see     #loadByValues
    */
   private Object[] collectRowData(int navigation)
   {
      if (cursor == null)
      {
         return null;
      }
      
      switch (navigation)
      {
         case NEXT:
         case LAST:
            if (cursor.isOffEnd(OffEnd.BACK))
            {
               return null;
            }
            break;
         case PREVIOUS:
         case FIRST:
            if (cursor.isOffEnd(OffEnd.FRONT))
            {
               return null;
            }
         default:
            break;
      }
      
      Object[] rowData = new Object[tableCount];
      int dstIdx = 0;
      Iterator<Component> iter = components.iterator();
      while (iter.hasNext())
      {
         Joinable query = iter.next().getQuery();
         Object[] subData = query.getRow();
         System.arraycopy(subData, 0, rowData, dstIdx, subData.length);
         dstIdx += query.getTableCount();
      }
      
      return rowData;
   }
   
   /**
    * Retrieve a virtual, composite row of results for the compound query.
    * This method implements a client-side join of multiple, single-row
    * record retrievals performed by the underlying, component queries.
    * It works first from outermost component inward, retrieving a record at
    * each level.  When each component's backing buffer contains a record,
    * the composite retrieval is complete and the method returns.  On
    * subsequent iterations, the innermost component is advanced first.  If
    * no more matches are possible at this level, the algorithm moves outward
    * one level and attempts to advance that level.  On success, it moves
    * inward again, on failure, it moves outward again.  This continues until
    * either a new composite row of results is assembled, or no further
    * records can be found.  This progression ensures that each inner query
    * cycles through all of its matches before the next outer query advances.
    *
    * @param   navigation
    *          Type of advancement applied to the compound query;  i.e., how
    *          the virtual composite result set is navigated as a whole:
    *          forward (next), backward (previous), jump to first or last, or
    *          reload current records in all buffers.
    * @param   lockType
    *          Overridding lock type to apply to each underlying buffer.
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @return  An array of the primary key IDs for every table involved in
    *          the query, (which together identify a unique, virtual record),
    *          if a result row was successfully retrieved;
    *          <code>null</code> if no more results were available (and we
    *          were not iterating).
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an error occurred retrieving the record for an underlying
    *          query.
    */
   private Object[] retrieveImpl(int navigation,
                                 LockType lockType,
                                 boolean iterating)
   {
      int size = components.size();
      int nullOuter = size;
      
      // If we are getting the first or last compound row, reset activeIndex
      // so that we are forced to get all components.  This prevents interim
      // state from interfering with a first/last request.
      switch (navigation)
      {
         case FIRST:
         case LAST:
            activeIndex = -1;
            break;
         default:
            break;
      }
      
      // Is this the first iteration in a new series?
      if (activeIndex < 0)
      {
         activeIndex = 0;
         
         // Create array to hold current loop counts, each initialized to 0.
         counters = new int[size];
         
         // Disable outer component queries' buffers' release-on-iterate
         // default behavior for the enclosing scope.  This is necessary to
         // prevent the premature release of an outer buffer's current
         // record while we are manually joining it with an inner buffer's
         // records.
         int end = size - 1;
         for (int i = 0; i < end; i++)
         {
            Component comp = components.get(i);
            RecordBuffer[] buffers = comp.getQuery().getRecordBuffers();
            int len = buffers.length;
            for (int j = 0; j < len; j++)
            {
               buffers[j].disableReleaseOnIterate();
            }
         }
      }
      
      boolean offEnd = false;
      for (int i = activeIndex; i < size; activeIndex = i)
      {
         Component comp = components.get(i);
         try
         {
            processComponent(navigation, comp, counters[i], lockType);
            counters[i]++;
            i++;
         }
         catch (QueryOffEndException exc)
         {
            if (comp.isOuter() && counters[i] == 0)
            {
               if (nullOuter == size)
               {
                  nullOuter = i;
               }
               comp.getQuery().setUnknownRecord();
               i++;
            }
            else
            {
               // Reset counter for this component.
               counters[i] = 0;
               
               // If we've propagated back up to the left-most component and
               // it has no more results, we're done.
               offEnd = (--i < 0);
               if (offEnd)
               {
                  if (iterating)
                  {
                     // If iterating, this is the only way out of the loop.
                     throw exc;
                  }
                  else
                  {
                     // If not iterating, assume client code is relying on
                     // an availability test to end loop.
                     break;
                  }
               }
            }
         }
         catch (ErrorConditionException exc)
         {
            activeIndex = -1;
            
            throw exc;
         }
      }
      
      activeIndex = Math.min(activeIndex, nullOuter) - 1;
      
      if (LOG.isTraceEnabled())
      {
         LOG.trace("activeIndex = " + activeIndex);
         LOG.trace("nullOuter = " + nullOuter);
         StringBuffer buf = new StringBuffer("counters:  ");
         for (int i = 0; i < size; i++)
         {
            buf.append(counters[i]);
            buf.append(" ");
         }
         LOG.trace(buf.toString());
      }
      
      if (!offEnd)
      {
         if (!postponeAccumulation)
         {
            // Notify accumulators of a successful iteration.
            accumulate();
         }
         
         incrementMoves();
      }
      
      return (offEnd ? null : collectRowData(navigation));
   }
   
   /**
    * Switch on postponed accumulation on next retrieve.
    */
   protected void postponeAccumulation()
   {
      postponeAccumulation = true;
   }
   
   /**
    * Switch on postponed accumulation on next retrieve.
    */
   protected void restoreAccumulation()
   {
      postponeAccumulation = false;
   }
   
   /**
    * Advance a single query component, according to the overall query
    * navigation direction, and that component's natural iteration type.  It
    * is the intersection of these two criteria which determines the actual
    * type of retrieval which is performed.
    *
    * @param   navigation
    *          Type of advancement applied to the compound query;  i.e., how
    *          the virtual composite result set is navigated as a whole:
    *          forward (next), backward (previous), jump to first or last, or
    *          reload current records in all buffers.
    * @param   component
    *          Query component being asked to retrieve a record.
    * @param   count
    *          Index of this query component's result within the current join.
    * @param   lockType
    *          Lock type to apply to the retrieved record for this query
    *          component
    *
    * @throws  QueryOffEndException
    *          if query component is being asked to retrieve a FIRST, LAST,
    *          or UNIQUE record, and it has already done so in the current
    *          iteration pass (prevents looping infinitely on these
    *          constructs).
    *
    * @see     QueryConstants#RETRIEVE_MODES
    */
   private void processComponent(int navigation,
                                 Component component,
                                 int count,
                                 LockType lockType)
   {
      int iteration = component.getIteration();
      int retrieveMode = RETRIEVE_MODES[navigation][iteration];
      
      // Trigger end condition for one-shot iteration types if iterating and
      // we've already retrieved a record for this component in this iteration
      // pass.
      switch (retrieveMode)
      {
         case FIRST:
         case LAST:
         case UNIQUE:
            if (count > 0)
            {
               throw new QueryOffEndException("", 
                    getOffEndListeners().toArray(new QueryOffEndListener[0]));
            }
            break;
            
         default:
            break;
      }
      
      Joinable query = component.getQuery();
      
      // Force the query to clear its state at the beginning of a loop pass.
      if (count == 0)
      {
         query.reset(!resolveOnce);
      }
      
      // Perform component query record retrieval.
      switch (retrieveMode)
      {
         case FIRST:
            query.first(lockType);
            break;
           
         case LAST:
            query.last(lockType);
            break;
            
         case NEXT:
            query.next(lockType);
            break;
            
         case PREVIOUS:
            query.previous(lockType);
            break;
            
         case UNIQUE:
            query.unique(lockType);
            break;
            
         case CURRENT:
            query.current(lockType);
            break;
            
         default:
            throw new UnsupportedOperationException(
               "Unknown record retrieval mode:  " + retrieveMode);
      }
   }
   
   /**
    * A helper container for a query component. A query component defines
    * the <code>Joinable</code> object to which record retrieval tasks
    * are delegated, the natural iteration type of this component query, and
    * the type of join to perform with the previous component (if any).
    */
   private static final class Component
   {
      /** Component query */
      private final Joinable query;
      
      /** Iteration behavior (from {@link QueryConstants}) */
      private final int iteration;
      
      /** <code>true</code> for outer join;  <code>false</code> for inner */
      private boolean outer;
      
      /**
       * Constructor.
       *
       * @param   query
       *          Component query.
       * @param   iteration
       *          Iteration behavior (from {@link QueryConstants}).
       * @param   outer
       *          <code>true</code> for outer join;  <code>false</code> for
       *          inner.
       */
      Component(Joinable query, int iteration, boolean outer)
      {
         this.query = query;
         this.iteration = iteration;
         this.outer = outer;
         
         for (RecordBuffer buffer : query.getRecordBuffers())
         {
            buffer.initialize();
         }
      }
      
      /**
       * Get the delegate query.
       *
       * @return  Query.
       */
      Joinable getQuery()
      {
         return query;
      }
      
      /**
       * Get the natural iteration type.
       *
       * @return  Iteration constant.
       */
      int getIteration()
      {
         return iteration;
      }
      
      /**
       * Does this query join to the previous query component using a left
       * outer join?
       *
       * @return  <code>true</code> if left outer join;  <code>false</code>
       *          if inner join.
       */
      boolean isOuter()
      {
         return outer;
      }
      
      /**
       * Disable the left outer join designation of this query component.
       * Exposed to allow outer join components to be reset to inner join
       * components, if an inner join component is added after an outer join
       * component has been added.
       */
      void setInner()
      {
         this.outer = false;
      }
   }
}
